#include "defs.h"
#include "xparameters.h"
#include "cache.h"
#include "Zynq7000_intc.h"
#include "cdma.h"

volatile static int Done = 0;  // DMA transfer is done
volatile static int Error = 0; // DMA Bus Error occurred
static UINT32 destp;
static int destlen;

static void CDMA_IRQHandler(int cdmaNum);
static void CDMA0_IRQHandler(void* param);
  
CDMA_PORT_T CDMA_Port[TOTAL_CDMA_PORT] = {
  { (volatile CDMA_Type *)XPAR_SPDIF_AXI_CDMA_0_BASEADDR, XPAR_FABRIC_SPDIF_AXI_CDMA_0_CDMA_INTROUT_INTR, CDMA0_IRQHandler, FALSE }  // CDMA0
};
  
static void CDMA0_IRQHandler(void* param) { CDMA_IRQHandler(CDMA0); }

static void CDMA_IRQHandler(int cdmaNum) { 
  UINT32 status, irq;
  volatile CDMA_Type *cdma = CDMA_REG(cdmaNum); 

  status = cdma->SR;
  if ((irq = status & XAXICDMA_XR_IRQ_ALL_MASK) == 0)
    return; // No interrupt for intr handler
  
  cdma->SR = irq;  // acknowledge the interrupt
  CDMA_SimpleNotDone(cdmaNum) = FALSE;
 
  // if has error interrupt, hardware needs to be reset
  if ((irq & XAXICDMA_XR_IRQ_ERROR_MASK) && (status & XAXICDMA_SR_ERR_ALL_MASK)) {
    Error = TRUE;
    cdma->CR = XAXICDMA_CR_RESET_MASK;
    for (int timeout = XAXICDMA_RESET_LOOP_LIMIT; timeout > 0; timeout--) {
      if (!(cdma->CR & XAXICDMA_CR_RESET_MASK))
        break;
    }
  }
  Done = TRUE;
  //if (!Error)
    // Invalidate the DestBuffer before receiving the data, in case the Data Cache is enabled
    //Xil_DCacheInvalidateRange(destp, destlen);

  return;
}

int CDMA_init(int cdmaNum) {
  int timeout, err;
  volatile CDMA_Type *cdma = CDMA_REG(cdmaNum); 

  for (;;) {
    err = RET_ERR;

    if ((((unsigned int)XPAR_SPDIF_AXI_CDMA_0_M_AXI_DATA_WIDTH) >> 3) < 4)  // AXI CDMA supports 32 bits data width and up
      break; // word length too short

    CDMA_SimpleNotDone(cdmaNum) = FALSE;  // mark no outstanding transfers

    // Reset the hardware
    cdma->CR = XAXICDMA_CR_RESET_MASK;
    for (timeout = XAXICDMA_RESET_LOOP_LIMIT; timeout > 0; timeout--) {
      if (!(cdma->CR & XAXICDMA_CR_RESET_MASK))
        break;
    }
    if (timeout == 0)
      break; // reset failed

    CPU_INTC_ActivateInterrupt(CDMA_IRQ(cdmaNum), CDMA_ISR(cdmaNum), 0);
    cdma->CR |= XAXICDMA_XR_IRQ_ALL_MASK;  // enable all (completion/error/delay) interrupts

    err = RET_OK;
    break;  // exit the forever loop
  }
  return err;
}

int CDMA_SimpleTransfer(int cdmaNum, UINT32 srcAddr, UINT32 dstAddr, int length) {
  UINT32 WordBits;
  int err;
  volatile CDMA_Type *cdma = CDMA_REG(cdmaNum); 

  for (;;) {
    err = RET_ERR;

    // save a copy of destination address and length in order to invalidate the cache in the isr
    destp = dstAddr;
    destlen = length;

    // flush the srcBuffer before the DMA transfer, in case the Data Cache is enabled
    Xil_DCacheFlushRange(srcAddr, length);

    Done = Error = 0;
    if ((length < 1) || (length > XAXICDMA_MAX_TRANSFER_LEN))
      break;

    WordBits = (UINT32)(((unsigned int)XPAR_SPDIF_AXI_CDMA_0_M_AXI_DATA_WIDTH) >> 3 - 1);

    if ((srcAddr & WordBits) || (dstAddr & WordBits)) {
      if (XPAR_SPDIF_AXI_CDMA_0_INCLUDE_DRE == 0)
        break;  // unaligned transfer without DRE
    }
    // cannot submit if the engine is doing a transfer
    if (!(cdma->SR & XAXICDMA_SR_IDLE_MASK))
      break;  // engine is busy
 
    // the driver is still handling the previous simple transfer
    if (CDMA_SimpleNotDone(cdmaNum) == TRUE)
      break;  // Simple ongoing
 
    // setup the flag so that others will not step on us (only set if the system is in interrupt mode)
    if ((cdma->CR & XAXICDMA_XR_IRQ_SIMPLE_ALL_MASK) != 0x0)
      CDMA_SimpleNotDone(cdmaNum) = TRUE;

    cdma->SRCADDR = srcAddr;
    cdma->DSTADDR = dstAddr;
    cdma->BTT = length;  // writing to the BTT register starts the transfer

    err = RET_OK;
    break;  // exit the forever loop
  }
  return err;
}
