#include <FreeRTOS.h>
#include <semphr.h>

#include "defs.h"
#include "xparameters.h"  // Zynq parameters
#include "Zynq7000_intc.h"
#include "fifo.h"
#include "cdma.h"

#define WORD_SIZE 4	    // fifo: size of words in bytes

static int errorCounter;  // FIFO error counter set in isr

extern SemaphoreHandle_t fifoSemaphore;

extern int activeBuf;  // switch between buf0 and buf1
extern short *shortBuf;
extern short *shortBuf0;
extern short *shortBuf1;
extern int exitPlay;

// fpga bram blocks
static UINT32 bram0 = XPAR_BRAM_0_BASEADDR;
static UINT32 bram1 = XPAR_BRAM_1_BASEADDR;

static void FIFO_IRQHandler(int fifoNum);
static void FIFO0_IRQHandler(void* param) { FIFO_IRQHandler(0); }

FIFO_PORT_T  FIFO_Port[TOTAL_FIFO_PORT] = {
  { (volatile FIFO_Type *)XPAR_AXI_FIFO_0_BASEADDR, XPAR_FABRIC_SPDIF_AXI_FIFO_MM_S_0_INTERRUPT_INTR, FIFO0_IRQHandler } // FIFO0
};


// bram0/bram1 = 64KB (=16K 64bit)
// a19/a18/a17/a16 (= bit3/bit2/bit1/bit0)
// 0   0   0   0	bram0 part1
// 0   0   0   1	bram1 "
// 0   0   1   0	bram0 "
// 0   0   1   1	bram1 "
// 0   1   0   0	bram0 "
// 0   1   0   1	bram1 "
// 0   1   1   0	bram0 "
// 0   1   1   1	bram1 "
// ---------------------------------------------------------------------------
static void FIFO_IRQHandler(int fifoNum) {
  volatile FIFO_Type *fifo = FIFO_REG(fifoNum);
  uint32 ReceiveLength, Pending;
  uint32 flip;
  BaseType_t xHigherPriorityTaskWoken; 

  Pending = fifo->IER & fifo->ISR;
  while (Pending) {
    if (Pending & XLLF_INT_RC_MASK) {
      ReceiveLength = fifo->RLF/WORD_SIZE;
      for (int i = 0; i < ReceiveLength; i++)
        flip = fifo->RDFD;
      short *src = activeBuf == 0 ? shortBuf1 : shortBuf0;

      int len = (NB_FRAMES*SAMPLES_PER_FRAME)/4;
      src += (len * (flip & 7)); 
      if ((flip & 7) == 7) {
        xSemaphoreGiveFromISR(fifoSemaphore, &xHigherPriorityTaskWoken);
        portYIELD_FROM_ISR(xHigherPriorityTaskWoken);
      }
      if (exitPlay == TRUE)
        src = shortBuf;  // empty buffer
      CDMA_SimpleTransfer(CDMA0, (UINT32)src, (flip & 1) == 1 ? bram0 : bram1, len*sizeof(short));
      fifo->ISR = XLLF_INT_RC_MASK & XLLF_INT_ALL_MASK;
    }
    else if (Pending & XLLF_INT_ERROR_MASK){
      if (Pending & XLLF_INT_RPURE_MASK) {
        fifo->RDFR = XLLF_RDFR_RESET_MASK;
      } else if (Pending & XLLF_INT_RPORE_MASK) {
        fifo->RDFR = XLLF_RDFR_RESET_MASK;
      } else if(Pending & XLLF_INT_RPUE_MASK) {
        fifo->RDFR = XLLF_RDFR_RESET_MASK;
      } else if (Pending & XLLF_INT_TPOE_MASK) {
        fifo->TDFR = XLLF_TDFR_RESET_MASK;
      } else if (Pending & XLLF_INT_TSE_MASK) {
      } 
      errorCounter++;
      fifo->ISR = XLLF_INT_ERROR_MASK & XLLF_INT_ALL_MASK;
    }
    else {
      fifo->ISR =  Pending & XLLF_INT_ALL_MASK;
    }
    Pending = fifo->IER & fifo->ISR;
  }
}

// ---------------------------------------------------------------------------
int FifoInit(int fifoNum) {
  volatile FIFO_Type *fifo = FIFO_REG(fifoNum);
  int status;

  errorCounter = 0;

  fifo->TDFR = XLLF_TDFR_RESET_MASK;
  fifo->RDFR = XLLF_RDFR_RESET_MASK;

  // Reset the core and generate the external reset by writing to
  // the Local Link/AXI Streaming Reset Register
  fifo->LLR = XLLF_RDFR_RESET_MASK;
 
  // Check for the Reset value
  status = fifo->ISR;
  fifo->ISR = XLLF_INT_ALL_MASK;
  status = fifo->ISR;
  if(status != 0)
    return RET_ERR;
    
  CPU_INTC_ActivateInterrupt(FIFO_IRQ(fifoNum), FIFO_ISR(fifoNum), 0);
  //fifo->IER |= XLLF_INT_ALL_MASK;

  return RET_OK;
}
