﻿using System;
using System.Threading;
using System.Collections;
using System.IO;
using Microsoft.SPOT;
using Microsoft.SPOT.Input;
using Microsoft.SPOT.Presentation.Controls;
using Microsoft.SPOT.Hardware;
using Zynq;

namespace CDplayer
{
    public class CmdEventArgs : EventArgs
    {
        byte cmd;
        string name = "";
        int fpos = 0;

        public byte Cmd { get { return cmd; } set { cmd = value; } }
        public string Name { get { return name; } set { name = value; } }
        public int Fpos { get { return fpos; } set { fpos = value; } }

        public CmdEventArgs() { }
        public CmdEventArgs(byte cmd) { this.cmd = cmd; }
    }

    public class CD
    {
        public delegate void cdEvent(object sender, CDS_TYPE cdsType);
        public event cdEvent cd;

        const int MAX_TRACKS = 50;

        CDS_TYPE cdsType;
        Canvas canvas;
        string filename;
        string playingCD;
        int fpos;

        int track;
        int nbTracks;
        bool isPlaying;
        bool isActive;
        string rootDirectory;
        string curCue;
        int[] trackArray;

        Text trackInfo;
        Thread cpu1Thread;
        Queue cpu1Queue;
        AutoResetEvent cpu1AREvent;
        CmdEventArgs playEventArgs;
        static FunctionSlave functionSlave = new FunctionSlave();
        Slave.Irq fposUpdateIrq;

        public CD(Canvas canvas)
        {
            this.canvas = canvas;
            isPlaying = false;
            isActive = false;
            filename = "";
            curCue = "";
            playingCD = "";
            fpos = 0;
            track = 1;
            nbTracks = 0;
            trackArray = new int[MAX_TRACKS];
            trackInfo = new Text();
            trackInfo.Font = Resources.GetFont(Resources.FontResources.NinaB);
            playEventArgs = new CmdEventArgs((byte)'P');
            cpu1AREvent = new AutoResetEvent(false);
            cpu1Queue = new Queue();
            cpu1Thread = new Thread(cpu1);
            cpu1Thread.Start();
        }
        
        void cpu1()
        {
            ToCpu1 toCpu1;
            CmdEventArgs args;

            toCpu1 = new ToCpu1(functionSlave);
            fposUpdateIrq = Slave.Irq.ID1;
            InterruptSlave fposUpdate = new InterruptSlave(fposUpdateIrq);
            fposUpdate.OnInterrupt += new NativeEventHandler(fposUpdate_OnInterrupt);
            toCpu1.setIrq(fposUpdateIrq);

            for (; ; )
            {
                cpu1AREvent.WaitOne();
                if (cpu1Queue.Count != 0)
                {
                    args = (CmdEventArgs)cpu1Queue.Dequeue();
                    byte cmd = args.Cmd;
                    int fpos = args.Fpos;
                    string name = args.Name;
                    toCpu1.cdCommand(cmd, name, fpos);
                }
            }
        }
        
        void fposUpdate_OnInterrupt(uint data1, uint data2, DateTime time)
        {
            fpos = (int)data1;
            canvas.Dispatcher.BeginInvoke(new DispatcherOperationCallback(display), null);
        }

        public object display(object f)
        {
            display(filename);
            return null;
        } 

        public string PlayingCD { get { return playingCD; } set { playingCD = value; } }
        public string RootDirectory { set { rootDirectory = value; } }

        public void Display(CDS_TYPE cdsType, string filename)
        {
            this.cdsType = cdsType;  // save cds type to return to correct one afterward
            isActive = true;
            display(filename);
        }

        void display(string filename)
        {
            if (isActive)
            {
                try
                {
                    canvas.Children.Clear();
                    this.filename = filename;

                    string cuename = filename.Substring(filename.LastIndexOf("\\"));
                    parseCue(rootDirectory + "\\music" + cuename.Substring(0, cuename.Length - 3) + "cue");

                    string commonDirectory = filename.Substring(0, filename.IndexOf("\\", 1)) + "\\common\\";
                    int lastIndex = filename.LastIndexOf("\\");
                    string path = filename.Substring(0, lastIndex - 4);  // isolates the pathname less "\\icon"
                    string name = filename.Substring(lastIndex + 1);
                    addImage(path + "img\\" + name, 65, 10, true, play_TouchEvent);  // cd full size image

                    // add control buttons
                    addImage(commonDirectory + (isPlaying ? "pause.bmp" : "play.bmp"), 80, 210, false, play_TouchEvent);
                    addImage(commonDirectory + "mfolder1.bmp", 265, 5, false, mfolder_TouchEvent);

                    if (filename == playingCD)
                    {
                        addImage(commonDirectory + "bwd1.bmp", 8, 160, false, bwd_TouchEvent);
                        addImage(commonDirectory + "fwd1.bmp", 260, 160, false, fwd_TouchEvent);

                        int track = 1;
                        for (int i = 0; i < MAX_TRACKS; i++)
                        {
                            if ((trackArray[i + 1] == 0) || (fpos >= trackArray[i] && fpos < trackArray[i + 1]))
                            {
                                track = i + 1;
                                break;
                            }
                        }
                        int pos = fpos / (441 * 4);
                        int p1 = pos / 6000;
                        int p2 = (pos - (p1 * 6000)) / 100;
                        int p3 = pos - (p1 * 6000) - (p2 * 100);
                        trackInfo.TextContent = track.ToString() + ".  " + p1.ToString() + ":" + p2.ToString() + ":" + p3.ToString();
                        canvas.Children.Add(trackInfo);
                        Canvas.SetLeft(trackInfo, 130);
                        Canvas.SetTop(trackInfo, 215);
                    }
                }
                catch (Exception)
                {
                    cd(this, cdsType);
                }
            }
        }

        private void bwd_TouchEvent(object sender, TouchEventArgs e)
        {
            if (track > 1)
            {
                track--;
                fpos = trackArray[track - 1];
                if ((filename == playingCD) && isPlaying)
                {
                    string name = filename.Substring(filename.LastIndexOf("\\"));
                    playEventArgs.Name = "music" + name.Substring(0, name.Length - 3) + "wav";
                    playEventArgs.Fpos = fpos + 68;
                    cpu1AREvent.Set();
                    cpu1Queue.Enqueue(playEventArgs);
                }
                display(filename);
            }
        }

        private void fwd_TouchEvent(object sender, TouchEventArgs e)
        {
            if (track <= nbTracks)
            {
                track++;
                fpos = trackArray[track - 1];
                if ((filename == playingCD) && isPlaying)
                {
                    string name = filename.Substring(filename.LastIndexOf("\\"));
                    playEventArgs.Name = "music" + name.Substring(0, name.Length - 3) + "wav";
                    playEventArgs.Fpos = fpos + 68;
                    cpu1AREvent.Set();
                    cpu1Queue.Enqueue(playEventArgs);
                }
                display(filename);
            }
        }

        private void play_TouchEvent(object sender, TouchEventArgs e)
        {
            if (filename != playingCD)
            {
                // play filename, track
                playingCD = filename;
                string name = filename.Substring(filename.LastIndexOf("\\"));
                playEventArgs.Name = "music" + name.Substring(0, name.Length - 3) + "wav";
                track = 1;
                fpos = 0;
                playEventArgs.Fpos = fpos + 68;  // add wave header
                isPlaying = true;
            }
            else
            {
                // pause/resume play
                playEventArgs.Name = "";
                playEventArgs.Fpos = isPlaying ? 0 : -1;
                isPlaying = !isPlaying;
            }
            cpu1AREvent.Set();
            cpu1Queue.Enqueue(playEventArgs);

            display(filename);
        }

        private void mfolder_TouchEvent(object sender, TouchEventArgs e)
        {
            isActive = false;
            cd(this, cdsType);
        }

        private void parseCue(string name)
        {
            if (name != curCue)
            {
                FileStream fs = new FileStream(name, FileMode.Open, FileAccess.Read);
                byte[] data = new byte[fs.Length];
                int read_count = fs.Read(data, 0, data.Length);
                fs.Close();

                for (int i = 0; i < MAX_TRACKS; i++)
                {
                    trackArray[i] = 0;
                }
                nbTracks = 0;
                int idx;
                int pos;
                string line = "";
                int start = 0;
                for (int i = 0; i < data.Length; i++)
                {
                    if (data[i] == 0x0a)
                    {
                        line = new String(System.Text.Encoding.UTF8.GetChars(data, start, i - 1 - start));
                        start = i + 1;
                        if ((idx = line.IndexOf("  TRACK ", 0, 8)) != -1)
                        {
                            if ((nbTracks + 1) != Convert.ToInt32(line.Substring(8, 2)))
                                break;  // illegal value
                        }
                        else if ((idx = line.IndexOf("    INDEX 01 ", 0, 13)) != -1)
                        {
                            pos = Convert.ToInt32(line.Substring(13, 2)) * 6000;
                            pos += Convert.ToInt32(line.Substring(16, 2)) * 100;
                            pos += Convert.ToInt32(line.Substring(19, 2));
                            trackArray[nbTracks] = pos * (441 * 4);  // 441 x 16-bit stereo samples per 1/100 second
                            nbTracks++;
                        }
                    }
                }
                nbTracks--;
                curCue = name;
            }
        }

        void addImage(String filename, int left, int top, bool border, TouchEventHandler eventHandler)
        {
            FileStream fs = new FileStream(filename, FileMode.Open, FileAccess.Read);
            byte[] data = new byte[fs.Length];
            int read_count = fs.Read(data, 0, data.Length);
            fs.Close();

            Image img = new Image(new Bitmap(data, Bitmap.BitmapImageType.Bmp));
            if (border)
            {
                Border bd = new Border();
                bd.SetBorderThickness(1);
                bd.Child = img;
                canvas.Children.Add(bd);
                Canvas.SetLeft(bd, left);
                Canvas.SetTop(bd, top);
                if (eventHandler != null)
                    bd.TouchDown += eventHandler;
            }
            else
            {
                canvas.Children.Add(img);
                Canvas.SetLeft(img, left);
                Canvas.SetTop(img, top);
                if (eventHandler != null)
                    img.TouchDown += eventHandler;
            }
        }
    }
}
