﻿using System;
using System.IO;
using Microsoft.SPOT;
using Microsoft.SPOT.Input;
using Microsoft.SPOT.Presentation.Controls;
using Microsoft.SPOT.Presentation.Media;

namespace CDplayer
{
    public enum CDS_TYPE
    {
        FOLDER,
        LIST
    }

    public class CDS
    {
        const int SLIDER_TOP = 60;
        const int SLIDER_LEFT = 265;
        const int SLIDER_WIDTH = 50;
        const int SLIDER_HEIGHT = 178;

        // in the cdlist screen, the first icon allows to return to the cdfolder screen
        public delegate void cdFolderEvent(object sender);
        public event cdFolderEvent cdFolder = null;

        // if cdfolder screen then goes to cdlist screen else if cdlist screen then goes to cdplaying screen
        public delegate void cdListEvent(object sender, CDS_TYPE cdsType, string filename);
        public event cdListEvent cdList = null;
        
        // icon to go to the cdplaying screen
        public delegate void cdPlayingEvent(object sender, CDS_TYPE cdsType, string filename);
        public event cdPlayingEvent cdPlaying = null;

        CDS_TYPE cdsType;
        Canvas canvas;
        int rows, columns;  // number of icons to display
        int iconWidth, iconHeight;
        String rootDirectory;
        string[] icons;     // all the icon files in the selected directory
        Slider slider;      // scrolling slider
        int firstDisplayedIcon;
        string playingCD;

        public CDS(CDS_TYPE cdsType, Canvas canvas, int rows, int columns, int iconWidth, int iconHeight)
        {
            this.canvas = canvas;
            this.rows = rows;
            this.columns = columns;
            this.cdsType = cdsType;
            this.iconWidth = iconWidth;
            this.iconHeight = iconHeight;
            firstDisplayedIcon = 0;
        }

        public void Init(string dir)
        {
            rootDirectory = dir.Substring(0, dir.IndexOf("\\", 1)); // extract the root directory name
            icons = Directory.GetFiles(dir);    // get all the icon files in the selected directory
            slider = null;
            firstDisplayedIcon = 0;

            // calculate the scrolling slider cursor height and setup the slider
            int nbItems = rows * columns;
            if (icons.Length > nbItems)
            {
                int n = ((icons.Length - nbItems) / rows) + 1;
                if ((icons.Length - nbItems) % rows != 0)
                    n++;
                int cursorHeight = SLIDER_HEIGHT / n;
                Color sliderColor = ColorUtility.ColorFromRGB(247, 242, 180);
                Color cursorColor = ColorUtility.ColorFromRGB(95, 141, 68);
                slider = new Slider(canvas, SLIDER_LEFT, SLIDER_TOP, SLIDER_WIDTH, SLIDER_HEIGHT, cursorHeight, sliderColor, cursorColor);
                slider.Slide += new Slider.sliderEvent(slide);
            }
        }
        
        public string PlayingCD { get { return playingCD; } set { playingCD = value; } }

        void slide(object sender, bool dir)
        {
            firstDisplayedIcon += dir ? rows : -rows;  // next/previous icons row
            Display();
        }

        public void Display()
        {
            canvas.Children.Clear();
            int i = firstDisplayedIcon;
            for (int y = 0; y < rows && i < icons.Length; y++)
            {
                for (int x = 0; x < columns && i < icons.Length; x++)
                {
                    if (cdsType == CDS_TYPE.LIST && x == 0 && y == 0)
                    {   // icon to return to the root
                        addImage(rootDirectory + "\\common\\up1.bmp", x * iconWidth, y * iconHeight, true, selectedFolder_TouchEvent);
                    }
                    else
                    {   // cd icon
                        addImage(icons[i], x * iconWidth, y * iconHeight, true, cd_TouchEvent);
                        i++;
                    }
                }
            }
            if (PlayingCD != "")  // icon to switch to currently playing cd
                addImage(rootDirectory + "\\common\\playing1.bmp", 265, 5, false, playingCD_TouchEvent);

            if (slider != null)
            {
                slider.Display();
            }
        }

        private void selectedFolder_TouchEvent(object sender, TouchEventArgs e)
        {
            cdFolder(this);
        }

        private void cd_TouchEvent(object sender, TouchEventArgs e)
        {
            int x = (e.Touches[0].X / iconWidth);
            int y = (e.Touches[0].Y / iconHeight);
            if (x < columns && y < rows)
            {
                string dir;
                if (cdsType == CDS_TYPE.LIST)
                {
                    dir = icons[x + (y * rows) + firstDisplayedIcon - 1];    // -1 to ignore the first icon (the return to root one)
                }
                else
                {
                    string name = icons[x + (y * rows) + firstDisplayedIcon];
                    // strips the file extension and point to the icon directory
                    dir = name.Substring(0, name.Length - 4) + "\\icon";
                }
                cdList(this, cdsType, dir);
            }
        }

        private void playingCD_TouchEvent(object sender, TouchEventArgs e)
        {
            cdPlaying(this, cdsType, PlayingCD);
        }

        void addImage(String filename, int left, int top, bool border, TouchEventHandler eventHandler)
        {
            FileStream fs = new FileStream(filename, FileMode.Open, FileAccess.Read);
            byte[] data = new byte[fs.Length];
            int read_count = fs.Read(data, 0, data.Length);
            fs.Close();

            Image img = new Image(new Bitmap(data, Bitmap.BitmapImageType.Bmp));
            if (border)
            {
                Border bd = new Border();
                bd.SetBorderThickness(1);
                bd.Child = img;
                canvas.Children.Add(bd);
                Canvas.SetLeft(bd, left);
                Canvas.SetTop(bd, top);
                if (eventHandler != null)
                    bd.TouchDown += eventHandler;
            }
            else
            {
                canvas.Children.Add(img);
                Canvas.SetLeft(img, left);
                Canvas.SetTop(img, top);
                if (eventHandler != null)
                    img.TouchDown += eventHandler;
            }
        }
    }
}
